#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 ~ 2012 Deepin, Inc.
#               2011 ~ 2012 Hou Shaohui
# 
# Author:     Hou Shaohui <houshao55@gmail.com>
# Maintainer: Hou Shaohui <houshao55@gmail.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from constant import PROGRAM_NAME

homedir = os.path.expanduser("~")
lastdir = homedir

data_home = os.environ.get('XDG_DATA_HOME') or \
            os.path.join(homedir, '.local', 'share')

config_home = os.environ.get('XDG_CONFIG_HOME') or \
            os.path.join(homedir, '.config')

cache_home = os.environ.get('XDG_CACHE_HOME') or \
            os.path.join(homedir, '.cache')

data_home = os.path.join(data_home, PROGRAM_NAME)
config_home = os.path.join(config_home, PROGRAM_NAME)
cache_home = os.path.join(cache_home, PROGRAM_NAME)

program_dir = os.path.split(os.path.dirname(os.path.realpath(__file__)))[0]

local_hack = False
# Detect if Exaile is not installed.
if os.path.exists(os.path.join(program_dir, 'setup.py')):
    local_hack = True

def get_config_dir():
    return config_home

def get_data_dir():
    return data_home

def get_cache_dir():
    return cache_home

def _get_path(basedirs, *subpath_elements, **kwargs):
    check_exists = kwargs.get("check_exists", True)
    subpath = os.path.join(*subpath_elements)
    for d in basedirs:
        path = os.path.join(d, subpath)
        if not check_exists or os.path.exists(path):
            return path
    return None

def get_last_dir():
    return lastdir

def get_specify_cache_dir(*subpath):
    path = os.path.join(get_cache_dir(), *subpath)
    if not os.path.exists(path):
        os.makedirs(path)
    return path    

def make_missing_dirs():
    if not os.path.exists(data_home):
        os.makedirs(data_home)
    if not os.path.exists(config_home):
        os.makedirs(config_home)
    if not os.path.exists(cache_home):
        os.makedirs(cache_home)
        
make_missing_dirs()        



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 ~ 2012 Deepin, Inc.
#               2011 ~ 2012 Hou Shaohui
# 
# Author:     Hou Shaohui <houshao55@gmail.com>
# Maintainer: Hou Shaohui <houshao55@gmail.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
    A generic framework for service providers, recommended to be used
    whenever there are multiple ways of accomplishing a task or multiple
    sources can offer the required data.
"""

from .events import event_manager

class ProviderManager(object):
    """
        The overall manager for services and providers for them
    """
    def __init__(self):
        self.services = {}

    def register(self, servicename, provider, target=None):
        """
            Registers a provider for a service. The provider object is used 
            by consumers of the service. 

            Services can be targeted for a specific use. For example, if you
            have a widget that uses a service 'foo', if your object can perform
            a service only for a specific type of widget, then target would be 
            set to the widget type. 
            
            If you had a service that could perform 'foo' for all widgets, then
            target would be set to None, and all widgets could use your service.
            
            It is intended that most services should set target to None, with 
            some narrow exceptions. 
            
            :param servicename: the name of the service [string]
            :type servicename: string
            :param provider: the object that is the provider [object]
            :type provider: object
            :param target: a specific target for the service [object]
            :type target: object
        """
        service = self.services.setdefault(servicename, {})
        providers = service.setdefault(target, [])
        if provider not in providers:
            providers.append(provider)
            event_manager.emit("%s_provider_added" % servicename, (provider, target), self)

    def unregister(self, servicename, provider, target=None):
        """
            Unregisters a provider.

            :param servicename: the name of the service
            :type servicename: string
            :param provider: the provider to be removed
            :type provider: object
            :param target: a specific target for the service [object]
            :type target: object
        """
        if servicename not in self.services:
            return
        try:
            service = self.services[servicename]
            if provider in service[target]:
                service[target].remove(provider)
                event_manager.emit("%s_provider_removed" % servicename, (provider, target), self)
                if not service[target]: #no values for target key then del it
                    del service[target]
        except KeyError:
            return

    def get(self, servicename, target=None):
        """
            Returns a list of providers for the specified servicename.

            This will return providers targeted for a specific target AND
            providers not targeted towards any particular target.
            
            :param servicename: the service name to get providers for
            :type servicename: string
            :param target: the target of the service
            :type target: object
            :returns: list of providers
            :rtype: list of objects
        """
        try:
            service = self.services[servicename]
        except KeyError:
            return []
        
        try:
            any = service[None]
        except KeyError:
            any = []
            
        if target is None:
            return any[:]
        
        try:
            specific = service[target]
        except KeyError:
            specific = []
        
        return specific + any

    def get_provider(self, servicename, providername, target=None):
        """
            Returns a single identified provider
            
            This will return a provider either targeted for the specific 
            target or a provider not targeted towards any particular target.

            :param servicename: The service name to get the provider for
            :type servicename: string
            :param providername: The provider name to identify the provider
            :type providername: string
            :param target: the target of the service
            :type target: object
            :returns: a provider or None
            :rtype: object
        """
        for provider in self.get_providers(servicename, target):
            if provider.name == providername:
                return provider

        return None

provider_manager = ProviderManager()

class ProviderHandler(object):
    """
        Base class to handle providers
        for one specific service including
        notification about (un)registration
    """
    def __init__(self, servicename, target=None, simple_init=False):
        """
            Target is the object that the service is being performed for.
            Often, if the service is truly global and it doesn't make sense
            to target a service at a particular consumer, it can be None. 
        
            :param servicename: the name of the service to handle
            :type servicename: string
            :param target: the target for a provided service. Generally, 
            this will be the object that uses the service
            :type target: string
            :param simple_init: call on_provider_added for every element
                already registered on instantiation.
        """
        self.servicename = servicename
        self.target = target
        if simple_init:
            for provider in provider_manager.get(servicename, target):
                self.on_provider_added(provider)
        event_manager.connect("%s_provider_added" % servicename, self._add_callback)
        event_manager.connect("%s_provider_removed" % servicename, self._remove_callback)

    def _add_callback(self, obj, ptuple):
        """
            Mediator to call actual callback
            for added providers
        """
        provider, target = ptuple
        if target is None or target is self.target:
            self.on_provider_added(provider)

    def on_provider_added(self, provider):
        """
            Called when a new provider is added

            :param provider: the new provider
            :type provider: object
        """
        pass # for overriding

    def _remove_callback(self, obj, ptuple):
        """
            Mediator to call actual callback
            for removed providers
        """
        provider, target = ptuple
        if target is None or target is self.target:
            self.on_provider_removed(provider)

    def on_provider_removed(self, provider):
        """
            Called when a provider is removed

            :param provider: the removed provider
            :type provider: object
        """
        pass # for overriding

    def get_providers(self):
        """
            Returns a list of providers for this service

            :returns: list of providers
            :rtype: list of objects
        """
        return provider_manager.get(self.servicename, self.target)

    def get_provider(self, providername):
        """
            Returns a provider for this service.

            :param providername: The provider name to
                identify the provider
            :type providername: string
            :returns: A provider or None
            :rtype: object
        """
        return provider_manager.get_provider(self.servicename, providername, self.target)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 ~ 2012 Deepin, Inc.
#               2011 ~ 2012 Hou Shaohui
# 
# Author:     Hou Shaohui <houshao55@gmail.com>
# Maintainer: Hou Shaohui <houshao55@gmail.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import re

#levelno = logging.DEBUG
levelno = logging.INFO

def setLevelNo(n):
    global levelno
    levelno = ( 100 - (n * 10) )

logging.addLevelName(100, "DEPRECATED")
console_format = '%(levelname)-8s %(name)-30s %(message)s'
logging.basicConfig(level=levelno, format=console_format, datafmt="%H:%M:%S")


def objaddr(obj):
    string = object.__repr__(obj)
    m = re.search("at (0x\w+)",string)
    if m: return  m.group(1)[2:]
    return "       "


class Logger(object):

    def set_logname(self, name):
        self.__logname = name

    def get_logname(self):
        if hasattr(self,"__logname") and self.__logname :
            return self.__logname
        else:
            return "%s.%s"%(self.__module__,self.__class__.__name__)

    def logdebug(self, msg, *args, **kwargs): 
        mylogger = logging.getLogger(self.get_logname())
        mylogger.debug(msg, *args, **kwargs)

    def loginfo(self, msg, *args, **kwargs): 
        mylogger = logging.getLogger(self.get_logname())
        mylogger.info(msg, *args, **kwargs)

    def logwarn(self, msg, *args, **kwargs): 
        mylogger = logging.getLogger(self.get_logname())
        mylogger.warn(msg, *args, **kwargs)

    def logerror(self, msg, *args, **kwargs): 
        mylogger = logging.getLogger(self.get_logname())
        mylogger.error(msg, *args, **kwargs)

    def logcritical(self, msg, *args, **kwargs): 
        mylogger = logging.getLogger(self.get_logname())
        mylogger.critical(msg, *args, **kwargs)

    def logexception(self, msg, *args, **kwargs):
        mylogger = logging.getLogger(self.get_logname())
        mylogger.exception(msg, *args, **kwargs)

    def logdeprecated(self, msg, *args, **kwargs):
        mylogger = logging.getLogger(self.get_logname())
        mylogger.log(100,msg, *args, **kwargs)

def newLogger(name):
    l = Logger()
    l.set_logname(name)
    return l




#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 ~ 2013 Deepin, Inc.
#               2011 ~ 2013 Hou ShaoHui
# 
# Author:     Hou ShaoHui <houshao55@gmail.com>
# Maintainer: Hou ShaoHui <houshao55@gmail.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import ssl
import socket
import httplib
import urllib2

class VerifiedHTTPSConnection(httplib.HTTPSConnection):
    def connect(self):
        # overrides the version in httplib so that we do
        #    certificate verification
        sock = socket.create_connection((self.host, self.port), self.timeout)
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()
        # wrap the socket using verification with the root
        #    certs in trusted_root_certs
        self.sock = ssl.wrap_socket(sock,
                                    self.key_file,
                                    self.cert_file,
                                    cert_reqs=ssl.CERT_REQUIRED,
                                    ca_certs="trusted_root_certs")
        
        
        
# wraps https connections with ssl certificate verification
class VerifiedHTTPSHandler(urllib2.HTTPSHandler):
    '''
    >>> https_handler = VerifiedHTTPSHandler()
    >>> url_opener = urllib2.build_opener(https_handler)
    >>> handle = url_opener.open('https://www.example.com')
    >>> response = handle.readlines()
    >>> handle.close()    
    
    '''
    def __init__(self, connection_class = VerifiedHTTPSConnection):
        self.specialized_conn_class = connection_class
        urllib2.HTTPSHandler.__init__(self)
        
    def https_open(self, req):
        return self.do_open(self.specialized_conn_class, req)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 ~ 2012 Deepin, Inc.
#               2011 ~ 2012 Hou Shaohui
# 
# Author:     Hou Shaohui <houshao55@gmail.com>
# Maintainer: Hou Shaohui <houshao55@gmail.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import time
import threading
import urllib2
import socket
import traceback
import urlparse
from ftplib import FTP
socket.setdefaulttimeout(120)         # 2 minutes

from .logger import Logger
from .constant import BLOCK_SIZE
from .common import os_open
from .providers import provider_manager
from events import event_manager

std_headers = {
    'User-Agent': 'Mozilla/5.0 (Windows; U; Windows NT 6.1; '
        'en-US; rv:1.9.2) Gecko/20100115 Firefox/3.6',
    'Accept-Charset': 'ISO-8859-1,utf-8;q=0.7,*;q=0.7',
    'Accept': 'text/xml,application/xml,application/xhtml+xml,'
        'text/html;q=0.9,text/plain;q=0.8,image/png,*/*;q=0.5',
    'Accept-Language': 'en-us,en;q=0.5',
}


class BaseFetch(threading.Thread, Logger):
    
    def __init__(self, name, url, part_output_file, state_file, start_offset, conn_state, task):
        threading.Thread.__init__(self)
        self.setDaemon(True)
        self.name = name
        self.url = url
        self.part_output_file = part_output_file
        self.state_file = state_file
        self.start_offset = start_offset
        self.conn_state = conn_state
        self._task = task
        self.length = conn_state.chunks[name] - conn_state.progress[name]
        self.sleep_timer = 0
        self.need_to_quit = False
        self.need_to_sleep = False
        
    @staticmethod    
    def get_file_size(url):
        raise NotImplementedError
    
    @staticmethod
    def is_match(url):
        raise NotImplementedError

    def run(self):
        pass

class HTTPFetch(BaseFetch):

    def __init__(self, *args, **kwargs):
        BaseFetch.__init__(self, *args, **kwargs)
        
    @staticmethod    
    def get_file_size(url):
        try:
            conn = urllib2.urlopen(url, timeout=10)
            size = int(conn.info().getheaders("Content-Length")[0])
            conn.close()
            return size
        
        except urllib2.HTTPError:
            return 0
        
        except urllib2.URLError:
            return 0
        
        except ValueError:
            return 0
        
        except socket.timeout:
            return 0
        
        except Exception:
            traceback.print_exc(file=sys.stdout)
            return 0
        
    @staticmethod    
    def is_match(url):
        return urlparse.urlparse(url).scheme in  ["http", "https"]

    def run(self):
        if self.length == 0:
            return
        
        request = urllib2.Request(self.url, None)
        request.add_header('Range', 'bytes=%d-%d' % (self.start_offset,
                                                     self.start_offset + self.length))
        while True:
            try:
                data = urllib2.urlopen(request, timeout=6)
            except urllib2.URLError, u:
                error_info = "Connection %s: did not start with %s" % (self.name, u)
                self.logerror(error_info)
            else:
                break

        # Open the output file
        out_fd = os_open(self.part_output_file)
        os.lseek(out_fd, self.start_offset, os.SEEK_SET)

        #indicates if connection timed out on a try
        try_times = 5
        _error = ""
        while try_times > 0 and self.length > 0:
            if self.need_to_quit:
                return

            if self.need_to_sleep:
                time.sleep(self.sleep_timer)
                self.need_to_sleep = False

            if self.length >= BLOCK_SIZE:
                fetch_size = BLOCK_SIZE
            else:
                fetch_size = self.length
            try:
                data_block = data.read(fetch_size)
                
                if len(data_block) == 0:
                    self.loginfo( "Connection %s: [TESTING]: 0 sized block fetched.", self.name)
                    
                block_length = len(data_block)    
                # if len(data_block) != fetch_size:
                #     self.loginfo("Connection %s: len(data_block) != fetch_size, but continuing anyway.", self.name)
                #     self.run()
                #     return
            except socket.timeout, s:
                _error = "Connection %s timed out with %s" % (self.name, s)
                self.loginfo(_error)
                try_times -= 1
                continue
            except Exception, e:
                _error = "Connection %s Unknown error with %s" % (self.name, e)
                self.loginfo(_error)
                try_times -= 1
                continue

            self.length -= block_length
            
            self.conn_state.update_data_downloaded(block_length, int(self.name))
            os.write(out_fd, data_block)
            self.start_offset += len(data_block)
            self.conn_state.save_state(self.state_file)

        os.close(out_fd)    
        data.close()

        if try_times == 0:
            self._task.fetch_thread_error_update(self, _error)
        
class FTPFetch(BaseFetch):

    def __init__(self, *args, **kwargs):
        BaseFetch.__init__(self, *args, **kwargs)
        
    @staticmethod    
    def get_file_size(url):
        try:
            url_parse = urlparse.urlparse(url)
            ftp = FTP(url_parse.netloc)
            ftp.login()
            size = int(ftp.size(url_parse.path))
            ftp.quit()
            return size
        except Exception:
            traceback.print_exc(file=sys.stdout)
            return 0
            
    @staticmethod    
    def is_match(url):
        return urlparse.urlparse(url).scheme == "ftp"

    def run(self):
        if self.length == 0:
            return
        
        # Login.
        url_parse = urlparse.urlparse(self.url)
        ftp = FTP(url_parse.netloc)
        ftp.login()
        
        # Transfer data in binary mode.
        ftp.voidcmd("TYPE I")
        
        # Set offset.
        ftp.sendcmd("REST %s" % self.start_offset)
        
        # Start download.
        conn = ftp.transfercmd("RETR %s" % url_parse.path)
        
        # Open the output file
        out_fd = os_open(self.part_output_file)
        os.lseek(out_fd, self.start_offset, os.SEEK_SET)

        #indicates if connection timed out on a try
        try_times = 10
        _error = ""
        while try_times > 0 and self.length > 0:
            if self.need_to_quit:
                return

            if self.need_to_sleep:
                time.sleep(self.sleep_timer)
                self.need_to_sleep = False

            if self.length >= BLOCK_SIZE:
                fetch_size = BLOCK_SIZE
            else:
                fetch_size = self.length
                
            try:
                data_block = conn.recv(fetch_size)
                
                if len(data_block) == 0:
                    self.loginfo( "Connection %s: [TESTING]: 0 sized block fetched.", self.name)
                    
                block_length = len(data_block)    
            except socket.timeout, s:
                _error = s
                self.loginfo("Connection %s timed out with %s", self.name, s)
                try_times -= 1
                continue

            self.length -= block_length
            
            self.conn_state.update_data_downloaded(block_length, int(self.name))
            os.write(out_fd, data_block)
            self.start_offset += len(data_block)
            self.conn_state.save_state(self.state_file)
            
        # Clean work.    
        os.close(out_fd)    
        conn.close()

        if try_times == 0:
            error_info = "Connection %s error with %s" % (self.name, _error)
            self._task.emit("error", error_info, self._task)

provider_manager.register("fetch", HTTPFetch)        
provider_manager.register("fetch", FTPFetch)
        



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 ~ 2012 Deepin, Inc.
#               2011 ~ 2012 Hou Shaohui
# 
# Author:     Hou Shaohui <houshao55@gmail.com>
# Maintainer: Hou Shaohui <houshao55@gmail.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import time
import threading
import traceback
import Queue
from contextlib import contextmanager 

class FetchService(threading.Thread):
    
    def __init__(self, concurrent_thread_num=5):
        threading.Thread.__init__(self)
        
        self.setDaemon(True)
        
        self.concurrent_thread_num = concurrent_thread_num
        
        self.active_mission_list = []
        self.wait_mission_list = []
        self.mission_result_list = []
        
        self.thread_sync_lock = threading.Lock()
        self.mission_lock = Queue.Queue()
                
    def run(self):
        while True:
            result = self.mission_lock.get()
            self.start_missions(result)
            time.sleep(0.5)
   
    @contextmanager
    def sync(self):
        """
        Internal function do synchronize jobs.
        """
        self.thread_sync_lock.acquire()
        try:  
            yield  
        except Exception, e:  
            print 'function sync got error: %s' % e  
            traceback.print_exc(file=sys.stdout)
        else:  
            self.thread_sync_lock.release()
   
    def add_missions(self, missions):
        for mission in missions:
            mission.connect("pause", self.finish_missions)
            mission.connect("resume",  self.resume_missions)
            mission.connect("stop",  self.finish_missions)
            mission.connect("finish",  self.finish_missions)
        self.mission_lock.put(missions)    
            
    def resume_missions(self, mission, data):        
        mission.connect("pause", self.finish_missions)
        mission.connect("resume",  self.resume_missions)
        mission.connect("stop",  self.finish_missions)
        mission.connect("finish",  self.finish_missions)
        
        with self.sync_lock():
            if len(self.active_mission_list) >= self.concurrent_thread_num:
                self.wait_mission_list.insert(0, mission)
            else:    
                self.start_mission(mission)
            
    def start_missions(self, missions):
        for (index, mission) in enumerate(missions):
            # Add to wait list if active mission number reach max value.
            if len(self.active_mission_list) >= self.concurrent_thread_num:
                self.wait_mission_list += missions[index::]
                break
            
            # Otherwise start new mission.
            else:
                self.start_mission(mission)
                
    def start_mission(self, mission):            
        self.active_mission_list.append(mission)
        mission.start()
            
    def wake_up_wait_missions(self):        
        for mission in self.wait_mission_list:
            # Just break loop when active mission is bigger than max value.
            if len(self.active_mission_list) >= self.concurrent_thread_num:
                break
            # Otherwise add mission from wait list.
            else:
                # Remove from wait list.
                if mission in self.wait_mission_list:
                    self.wait_mission_list.remove(mission)
                
                # Start new mission.
                self.start_mission(mission)
        
    def finish_missions(self, mission,  data):        
        
        with self.sync():
            # Remove mission from active mission list.
            if mission in self.active_mission_list:
                mission.disconnect("pause", self.finish_missions)
                mission.disconnect("resume", self.resume_missions)
                mission.disconnect("stop", self.finish_missions)
                mission.disconnect("finish", self.finish_missions)
                self.active_mission_list.remove(mission)
                
            # Wake up wait missions.
            self.wake_up_wait_missions()
            
            # Exit thread when download finish.
            if (len(self.active_mission_list) == 0 and len(self.wait_mission_list) == 0):
                pass
                # self.mission_lock.put(self.FINISH_SIGNAL)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 ~ 2012 Deepin, Inc.
#               2011 ~ 2012 Hou Shaohui
# 
# Author:     Hou Shaohui <houshao55@gmail.com>
# Maintainer: Hou Shaohui <houshao55@gmail.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import math

def parse_bytes(bytes):
    if bytes == 0:
        return "0b"
    k = math.log(bytes, 1024)
    ret_str = "%.2f%s" % (bytes / (1024.0 ** int(k)), "bKMGTPEY"[int(k)])
    return ret_str

def parse_time(time_in_secs):
    ret_str = ""
    mult_list = [60, 60 * 60, 60 * 60 * 24]
    unit_list = ["second(s)", "minute(s)", "hour(s)", "day(s)"]
    for i in range(len(mult_list)):
        if time_in_secs < mult_list[i]:
            pval = int(time_in_secs / (mult_list[i - 1] if i > 0 else 1))
            ret_str = "%d %s" % (pval, unit_list[i])
            break
    if len(ret_str) == 0:
        ret_str = "%d %s" % (int(time_in_secs / mult_list[2]), \
                                  unit_list[3])
    return ret_str

class ProgressBar(object):
    
    def __init__(self, n_conn, conn_state):
        self.n_conn = n_conn
        self.dots = ["" for i in range(n_conn)]
        self.conn_state = conn_state

    def _get_term_width(self):
        platform = sys.platform
        default_cols = 75
        if platform.startswith("linux"):
            try:
                term_rows, term_cols = map(int, os.popen('stty size', 'r').read().split())
            except:    
                return default_cols
            else:
                return term_cols
        else:    
            return default_cols

    def _get_download_rate(self, bytes):
        ret_str = parse_bytes(bytes)
        ret_str += "/s."
        return len(ret_str), ret_str

    def _get_percentage_complete(self, dl_len):
        assert self.conn_state.filesize != 0
        ret_str = str(dl_len * 100 / self.conn_state.filesize) + "%."
        return len(ret_str), ret_str

    def _get_time_left(self, time_in_secs):
        ret_str = ""
        mult_list = [60, 60 * 60, 60 * 60 * 24]
        unit_list = ["second(s)", "minute(s)", "hour(s)", "day(s)"]
        for i in range(len(mult_list)):
            if time_in_secs < mult_list[i]:
                pval = int(time_in_secs / (mult_list[i - 1] if i > 0 else 1))
                ret_str = "%d %s" % (pval, unit_list[i])
                break
        if len(ret_str) == 0:
            ret_str = "%d %s." % (int(time_in_secs / mult_list[2]), \
                                      unit_list[3])
        return len(ret_str), ret_str

    def _get_pbar(self, width):
        ret_str = "["
        for i in range(self.n_conn):
            dots_list = ['=' for j in range((self.conn_state.progress[i] *
                                             width) /
                                            self.conn_state.chunks[i])]
            self.dots[i] = "".join(dots_list)
            if ret_str == "[":
                ret_str += self.dots[i]
            else:
                ret_str += "|" + self.dots[i]
            if len(self.dots[i]) < width:
                ret_str += '>'
                ret_str += "".join([' ' for i in range(width -
                                                       len(self.dots[i]) - 1)])

        ret_str += "]"
        return len(ret_str), ret_str

    def display_progress(self):
        dl_len = 0
        for rec in self.conn_state.progress:
            dl_len += rec

        try:
            avg_speed = dl_len / self.conn_state.elapsed_time
        except:    
            avg_speed = 0

        ldr, drate = self._get_download_rate(avg_speed)
        lpc, pcomp = self._get_percentage_complete(dl_len)
        ltl, tleft = self._get_time_left((self.conn_state.filesize - dl_len) /
                                         avg_speed if avg_speed > 0 else 0)
        # term_width - #(|) + #([) + #(]) + #(strings) +
        # 6 (for spaces and periods)
        available_width = self._get_term_width() - (ldr + lpc +
                                                        ltl) - self.n_conn - 1 - 6
        lpb, pbar = self._get_pbar(available_width / self.n_conn)
        sys.stdout.flush()
        print "\r%s %s %s %s" % (drate, pcomp, tleft, pbar),
            



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 ~ 2012 Deepin, Inc.
#               2011 ~ 2012 Hou Shaohui
# 
# Author:     Hou Shaohui <houshao55@gmail.com>
# Maintainer: Hou Shaohui <houshao55@gmail.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


__all__ = ["common", "events", "fetch", "logger", "report", "services", "state", "tasks", "xdg"]

from . import common
from . import events
from . import fetch
from . import logger
from . import report
from . import services
from . import state
from . import tasks
from . import xdg



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 ~ 2012 Deepin, Inc.
#               2011 ~ 2012 Hou Shaohui
# 
# Author:     Hou Shaohui <houshao55@gmail.com>
# Maintainer: Hou Shaohui <houshao55@gmail.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import time
import threading
import sys
import traceback

from .report import ProgressBar, parse_bytes
from .state import ConnectionState
from .fetch import provider_manager
from .events import EventManager

from . import common

class StopExcetion(Exception):
    pass

class PauseException(Exception):
    pass

class ResumeException(Exception):
    pass

class TaskObject(EventManager):
    
    def __init__(self, url, output_file=None, num_connections=4, max_speed=None,
            verbose=False, output_temp=False, update_interval=1, task_name=None):
        EventManager.__init__(self)

        self.url = url
        self.output_file = self.get_output_file(output_file)
        self.num_connections = num_connections
        self.task_name = common.get_task_id(url) if not task_name else task_name

        self.max_speed = max_speed
        self.conn_state = None
        self.fetch_threads = []
        self.__stop = True
        self.__pause = False
        self.__finish = False
        self.verbose = verbose
        self.output_temp = output_temp
        self.update_interval = update_interval
        self.update_object = common.Storage()
        self.task_thread = None
        
        self.RemoteFetch = None
        fetchs = provider_manager.get("fetch")
        
        for fetch in fetchs:        
            if fetch.is_match(url):
                self.RemoteFetch = fetch
                break

    def fetch_thread_error_update(self, thread, error_info):
        if thread in self.fetch_threads:
            self.fetch_threads.remove(thread)

        if len(self.fetch_threads) == 0:
            self.emit("error", "%s task occur error with %s" % (self.task_name, error_info), self)

    def get_task_id(self):
        pass
        
    def get_output_file(self, output_file):    
        if output_file is not None:
            return output_file
        
        return self.url.rsplit("/", 1)[1]
    
    def emit_update(self):
        dl_len = 0
        for rec in self.conn_state.progress:
            dl_len += rec
            
        try:    
            avg_speed = dl_len / self.conn_state.elapsed_time
        except:    
            avg_speed = 0
            
        self.update_object.speed = avg_speed    
        self.update_object.progress = dl_len * 100 / self.conn_state.filesize
        self.update_object.remaining = (self.conn_state.filesize - dl_len) / avg_speed if avg_speed > 0 else 0
        self.update_object.filesize = self.conn_state.filesize
        self.update_object.downloaded = dl_len
        
        self.emit("update", self.update_object, self)
        
    def is_actived(self):    
        for task in self.fetch_threads:
            if task.isAlive():
                return True
        return False    
    
    def stop_all_task(self):
        for task in self.fetch_threads:
            task.need_to_quit = True
            
    def stop(self):        
        self.__stop = True
        self.task_thread = None
        
    def pause(self):    
        self.__pause = True
        
    def resume(self):
        if self.task_thread is None:
            self.emit("resume", obj=self)
            self.start()
        
    def isfinish(self):    
        return self.__finish
    
    def start(self):
        self.task_thread = threading.Thread(target=self.run)
        self.task_thread.setDaemon(True)
        self.task_thread.start()
            
    def run(self):    
        try:
            if self.RemoteFetch is None:
                error_info = "Don't support the protocol"
                self.logerror(error_info)
                self.emit("error", error_info, self)
                return 
            
            if not self.output_file:
                error_info = "Invalid URL"
                self.logerror(error_info)
                self.emit("error", error_info, self)
                return

            
            self.__stop = False
            self.__pause = False
            
            file_size = self.RemoteFetch.get_file_size(self.url)
            if file_size == 0:
                error_info = "Failed to get file information"
                self.logerror("UEL: %s, %s", self.url, error_info)
                self.emit("error", error_info, self)
                return

            if os.path.exists(self.output_file):
                self.update_object.speed = -1
                self.update_object.progress = 100
                self.update_object.remaining = 0
                self.update_object.filesize = file_size
                self.update_object.downloaded = file_size
                self.emit("update", self.update_object, self)
                self.emit("finish", obj=self)
                return

            if self.output_temp:
                part_output_file = common.get_temp_file(self.url)
            else:
                part_output_file = "%s.part" % self.output_file
            
            self.emit("start", obj=self)
            
            # load ProgressBar.
            # if file_size < BLOCK_SIZE:
            #     num_connections = 1
            # else:    
            num_connections = self.num_connections
             
            # Checking if we have a partial download available and resume
            self.conn_state = ConnectionState(num_connections, file_size)    
            state_file = common.get_state_file(self.url)
            self.conn_state.resume_state(state_file, part_output_file)
            
            
            self.report_bar = ProgressBar(num_connections, self.conn_state)
            
            self.logdebug("File: %s, need to fetch %s", self.output_file, 
                         parse_bytes(self.conn_state.filesize - sum(self.conn_state.progress)))
            
            #create output file with a .part extension to indicate partial download

            part_output_file_fp = os.open(part_output_file, os.O_CREAT | os.O_WRONLY)
            os.close(part_output_file_fp)
            
            start_offset = 0
            start_time = time.time()

            for i in range(num_connections):
                current_thread = self.RemoteFetch(i, self.url, part_output_file, state_file, 
                                           start_offset + self.conn_state.progress[i],
                                           self.conn_state, self)
                self.fetch_threads.append(current_thread)
                current_thread.start()
                start_offset += self.conn_state.chunks[i]

            while self.is_actived():
                if self.__stop:
                    raise StopExcetion

                if self.__pause:
                    raise PauseException

                end_time = time.time()
                self.conn_state.update_time_taken(end_time-start_time)
                start_time = end_time

                download_sofar = self.conn_state.download_sofar()

                if self.max_speed != None and \
                        (download_sofar / self.conn_state.elapsed_time) > (self.max_speed * 1204):
                    for task in self.fetch_threads:
                        task.need_to_sleep = True
                        task.sleep_timer = download_sofar / (self.max_speed * 1024 - self.conn_state.elapsed_time)

                # update progress
                if self.verbose:
                    self.report_bar.display_progress()
                self.emit_update()
                time.sleep(self.update_interval)

            if len(self.fetch_threads) == 0:
                raise StopExcetion

            if self.verbose:
                self.report_bar.display_progress()

            try:
                os.remove(state_file)
                os.unlink(self.output_file)
            except:
                pass

            os.rename(part_output_file, self.output_file)
            self.__finish = True
            self.emit_update()
            self.emit("finish", obj=self)
            if self.verbose:
                self.report_bar.display_progress()

        except StopExcetion:
            self.stop_all_task()

            try:
                os.unlink(part_output_file)
            except: pass

            try:
                os.unlink(state_file)
            except: pass
            
            self.emit("stop", obj=self)
            
        except PauseException:    
            self.stop_all_task()
            self.emit("pause", obj=self)
            
        except KeyboardInterrupt, e:    
            self.emit("stop", obj=self)
            self.stop_all_task()
            
        except Exception, e:
            self.emit("error", "Unknown error", self)
            self.emit("stop", obj=self)
            traceback.print_exc(file=sys.stdout)
            self.logdebug("File: %s at dowloading error %s", self.output_file, e)
            self.stop_all_task()

class MultiTaskObject(EventManager):
    """
        make multi task like one task
    """
    def __init__(self, urls, output_dir=None, num_connections=2, 
            update_interval=1, task_name=None):
        EventManager.__init__(self)

        self.urls = urls
        self.output_dir = output_dir
        self.num_connections = num_connections
        self.update_interval = update_interval
        self.task_name = common.get_task_id(urls) if not task_name else task_name

        self.__stop = True
        self.__pause = False
        self.__finish = False
        self.task_thread = None
        self.start_time = None

        self.update_object = common.Storage()
        self.__downloaded_size = 0
        self.__last_downloaded = 0

        self.total_size = 0
        self.active_task_list = []
        self.wait_task_list = []
        self.all_task_list = []
        self.task_finish_list = []
        self.init_tasks()

    def init_tasks(self):
        for url in self.urls:
            task = TaskObject(url, update_interval=0.2)
            task.connect("update", self.update_tasks)
            task.connect("pause", self.pause_tasks)
            task.connect("stop", self.stop_tasks)
            task.connect("finish", self.finish_tasks)
            task.connect("resume", self.resume_tasks)
            task.connect("error", self.emit_error)
            self.all_task_list.append(task)

    def stop(self):
        self.__stop = True
        self.task_thread = None
        self.stop_all_task()

    def stop_all_task(self):
        for task in self.all_task_list:
            task.stop()

    def is_actived(self):
        return len(self.task_finish_list) != len(self.urls)

    def update_tasks(self, task, data):
        pass

    def pause_tasks(self, task, data):
        pass

    def stop_tasks(self, task, data):
        pass

    def finish_tasks(self, task, data):
        if task in self.active_task_list:
            task.disconnect("update", self.update_tasks)
            task.disconnect("pause", self.pause_tasks)
            task.disconnect("stop", self.stop_tasks)
            task.disconnect("finish", self.finish_tasks)
            task.disconnect("resume", self.resume_tasks)
            task.disconnect("error", self.emit_error)
            self.active_task_list.remove(task)
            if task not in self.task_finish_list:
                self.task_finish_list.append(task)
        self.wake_up_wait_tasks()

    def wake_up_wait_tasks(self):
        for task in self.wait_task_list:
            # Just break loop when active task is bigger than max value.
            if len(self.active_task_list) >= self.num_connections:
                break
            # Otherwise add task from wait list.
            else:
                # Remove from wait list.
                if task in self.wait_task_list:
                    self.wait_task_list.remove(task)
                    if task not in self.active_task_list:
                        self.active_task_list.append(task)
                        task.start()

    def resume_tasks(self, task, data):
        pass

    def emit_error(self, task, error_info):
        self.logerror(error_info)
        self.emit("error", error_info, self)
        self.__stop = True
        self.task_thread = None

    def emit_update(self):
        dl_len = 0
        valid_len = 0
        for task in self.all_task_list:
            if task.update_object:
                dl_len += task.update_object.downloaded
                if task.update_object.speed != -1:
                    valid_len += task.update_object.downloaded

        if self.__last_downloaded == 0:
            self.__last_downloaded = valid_len
        self.__downloaded_size += valid_len - self.__last_downloaded
        self.__last_downloaded = valid_len
        now = time.time()
        elapsed_time = now - self.start_time

        try:
            avg_speed = self.__downloaded_size / elapsed_time
        except:
            avg_speed = 0

        # TODO: speed
        self.update_object.speed = avg_speed
        self.update_object.progress = dl_len * 100 / self.total_size
        self.update_object.remaining = (self.total_size - dl_len) / avg_speed if avg_speed > 0 else 0
        self.update_object.filesize = self.total_size
        self.update_object.downloaded = dl_len

        self.emit("update", self.update_object, self)

    def start(self):
        self.task_thread = threading.Thread(target=self.run)
        self.task_thread.setDaemon(True)
        self.task_thread.start()

    def run(self):
        try:
            self.emit("start", obj=self)
            self.start_time = time.time()

            for (index, task) in enumerate(self.all_task_list):
                if task.RemoteFetch is None:
                    error_info = "Don't support the protocol"
                    self.logerror(error_info)
                    error_list = [error_info, task]
                    self.emit("error", error_list, self)
                    return

                if not task.output_file:
                    error_info = "Invalid URL"
                    self.logerror(error_info)
                    error_list = [error_info, task]
                    self.emit("error", error_list, self)
                    return

                file_size = task.RemoteFetch.get_file_size(task.url)
                if file_size == 0:
                    error_info = "Failed to get file information"
                    self.logerror("UEL: %s, %s", task.url, error_info)
                    error_list = [error_info, task]
                    self.emit("error", error_list, self)
                    return
                else:
                    self.total_size += file_size

                if self.output_dir:
                    task.output_file = os.path.join(self.output_dir, task.output_file)

                if len(self.active_task_list) >= self.num_connections:
                    self.wait_task_list += self.all_task_list[index:]
                else:
                    self.active_task_list.append(task)
                    task.start()

            self.__stop = False
            self.__pause = False

            while self.is_actived():
                if self.__stop:
                    raise StopExcetion

                if self.__pause:
                    raise PauseException

                if len(self.active_task_list) >= self.num_connections:
                    pass

                self.emit_update()
                time.sleep(self.update_interval)

            self.emit_update()
            self.emit("finish", obj=self)
            self.logdebug("%s task finished" % self.task_name)

        except StopExcetion:
            self.stop_all_task()
            self.emit("stop", obj=self)
            self.logdebug("%s task stopped" % self.task_name)

        except PauseException:
            self.stop_all_task()
            self.emit("pause", obj=self)
            self.logdebug("%s task paused" % self.task_name)

        except KeyboardInterrupt, e:
            self.emit("stop", obj=self)
            self.stop_all_task()
            self.logdebug("%s task stopped" % self.task_name)

        except Exception, e:
            self.stop_all_task()
            self.emit("error", "Unknown error", self)
            self.emit("stop", obj=self)
            self.logerror("MultiTask: %s at dowloading error %s", self.urls, e)
            traceback.print_exc(file=sys.stdout)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 ~ 2012 Deepin, Inc.
#               2011 ~ 2012 Hou Shaohui
# 
# Author:     Hou Shaohui <houshao55@gmail.com>
# Maintainer: Hou Shaohui <houshao55@gmail.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import cPickle
import hashlib

from .logger import newLogger
from .xdg import get_cache_dir, get_specify_cache_dir

logger = newLogger("common")

def get_md5(string):                
    return hashlib.md5(string).hexdigest()

def get_state_file(url):
    return os.path.join(get_cache_dir(), get_md5(url))

def get_temp_file(url):
    return os.path.join(get_specify_cache_dir("temp"), get_md5(url))

def get_task_id(obj):
    if isinstance(obj, list):
        md5s = [get_md5(url) for url in obj]
        return "%d_%s" % (len(obj), get_md5(".".join(md5s)))
    elif isinstance(obj, str):
        return get_md5(obj)

def save_db(objs, fn):
    '''Save object to db file.'''
    
    f = open(fn + ".tmp", "wb")
    cPickle.dump(objs, f, cPickle.HIGHEST_PROTOCOL)
    f.close()
    
    try:
        os.unlink(fn)
    except: pass    
    
    try:
        os.rename(fn + ".tmp", fn)
    except:    
        pass
    
def load_db(fn):    
    '''Load object from db file.'''
    
    objs = None
    
    if os.path.exists(fn):
        f = open(fn, "rb")
        try:
            objs = cPickle.load(f)
        except:    
            logger.logexception("%s is not a valid database.", fn)
            try:
                os.unlink(fn)
            except: pass    
            
            objs = None
        f.close()    
    return objs    


class Storage(dict):
    """
    A Storage object is like a dictionary except `obj.foo` can be used
    in addition to `obj['foo']`.
    
        >>> o = storage(a=1)
        >>> o.a
        1
        >>> o['a']
        1
        >>> o.a = 2
        >>> o['a']
        2
        >>> del o.a
        >>> o.a
        Traceback (most recent call last):
            ...
        AttributeError: 'a'
    
    """
    def __getattr__(self, key): 
        try:
            return self[key]
        except KeyError, k:
            raise AttributeError, k
    
    def __setattr__(self, key, value): 
        self[key] = value
    
    def __delattr__(self, key):
        try:
            del self[key]
        except KeyError, k:
            raise AttributeError, k
    
    def __repr__(self):     
        return '<Storage ' + dict.__repr__(self) + '>'
    
def os_open(filename):    
    platform = sys.platform
    if platform == "win32":
        flags = os.O_WRONLY | os.O_BINARY
    else:    
        flags = os.O_WRONLY
        
    return os.open(filename, flags)    



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 ~ 2012 Deepin, Inc.
#               2011 ~ 2012 Hou Shaohui
# 
# Author:     Hou Shaohui <houshao55@gmail.com>
# Maintainer: Hou Shaohui <houshao55@gmail.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from distutils.core import setup

setup(
    name            = 'deepin-storm',
    version         = '0.3',
    description     = 'Python download library and Powerful download manager',
    long_description= 'Deepin Storm is download library and powerful download manager',
    author          = 'Alan Leaf',
    author_email    = 'alan.leaf916@gmail.com',
    url             = 'https://github.com/linuxdeepin/deepin-storm',
    license         = 'GPL-3',
    platforms       = ['all'],
    packages        = ['deepin_storm']
    )



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 ~ 2012 Deepin, Inc.
#               2011 ~ 2012 Hou Shaohui
# 
# Author:     Hou Shaohui <houshao55@gmail.com>
# Maintainer: Hou Shaohui <houshao55@gmail.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from threading import Lock

from .logger import Logger
from . import common

class ConnectionState(Logger):
    
    def __init__(self, n_conn, filesize):
        self.n_conn = n_conn
        self.filesize = filesize
        self.progress = [0 for i in range(n_conn)]
        self.elapsed_time = 0
        self.chunks = [(filesize / n_conn) for i in range(n_conn)]
        self.chunks[0] += filesize % n_conn
        self.save_lock = Lock()
        
        self.save_objs = common.Storage()

    def download_sofar(self):
        dwnld_sofar = 0
        for rec in self.progress:
            dwnld_sofar += rec
        return dwnld_sofar

    def update_time_taken(self, elapsed_time):
        self.elapsed_time += elapsed_time

    def update_data_downloaded(self, fetch_size, conn_id):
        self.progress[conn_id] += fetch_size

    def resume_state(self, state_file, output_file):
        if not os.path.isfile(output_file):
            return 
        saved_obj = common.load_db(state_file)
        if not saved_obj:
            return 
        
        for p in "n_conn filesize progress chunks elapsed_time".split():
            if not hasattr(saved_obj, p):
                return 
        
        self.n_conn = saved_obj.n_conn
        self.filesize = saved_obj.filesize
        self.progress = saved_obj.progress
        self.chunks = saved_obj.chunks
        self.elapsed_time = saved_obj.elapsed_time
        
    def _save_state(self):    
        self.save_objs.n_conn = self.n_conn
        self.save_objs.filesize = self.filesize
        self.save_objs.progress = self.progress
        self.save_objs.chunks = self.chunks
        self.save_objs.elapsed_time = self.elapsed_time

    def save_state(self, state_file):
        #out_fd will be closed after save_state() is completed
        #to ensure that state is written onto the disk
        self.save_lock.acquire()
        self._save_state()
        common.save_db(self.save_objs, state_file)
        self.save_lock.release()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2011 ~ 2012 Deepin, Inc.
#               2011 ~ 2012 Hou Shaohui
# 
# Author:     Hou Shaohui <houshao55@gmail.com>
# Maintainer: Hou Shaohui <houshao55@gmail.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import with_statement

from inspect import ismethod
from new import instancemethod
import re
import sys
import threading
import time
import weakref
import traceback

from .logger import Logger

class Nothing(object):
    pass

_NONE = Nothing() # used by event for a safe None replacement

class _WeakMethod(object):
    """Represent a weak bound method, i.e. a method doesn't keep alive the
    object that it is bound to. It uses WeakRef which, used on its own,
    produces weak methods that are dead on creation, not very useful.
    Typically, you will use the getRef() function instead of using
    this class directly. """

    def __init__(self, method, notifyDead = None):
        """
            The method must be bound. notifyDead will be called when
            object that method is bound to dies.
        """
        assert ismethod(method)
        if method.im_self is None:
            raise ValueError, "We need a bound method!"
        if notifyDead is None:
            self.objRef = weakref.ref(method.im_self)
        else:
            self.objRef = weakref.ref(method.im_self, notifyDead)
        self.fun = method.im_func
        self.cls = method.im_class

    def __call__(self):
        if self.objRef() is None:
            return None
        else:
            return instancemethod(self.fun, self.objRef(), self.cls)

    def __eq__(self, method2):
        if not isinstance(method2, _WeakMethod):
            return False
        return      self.fun      is method2.fun \
                and self.objRef() is method2.objRef() \
                and self.objRef() is not None


    def __hash__(self):
        return hash(self.fun)

    def __repr__(self):
        dead = ''
        if self.objRef() is None:
            dead = '; DEAD'
        obj = '<%s at %s%s>' % (self.__class__, id(self), dead)
        return obj

    def refs(self, weakRef):
        """Return true if we are storing same object referred to by weakRef."""
        return self.objRef == weakRef

def _getWeakRef(obj, notifyDead=None):
    """
        Get a weak reference to obj. If obj is a bound method, a _WeakMethod
        object, that behaves like a WeakRef, is returned, if it is
        anything else a WeakRef is returned. If obj is an unbound method,
        a ValueError will be raised.
    """
    if ismethod(obj):
        createRef = _WeakMethod
    else:
        createRef = weakref.ref

    if notifyDead is None:
        return createRef(obj)
    else:
        return createRef(obj, notifyDead)

class Event(object):
    """
        Represents an Event
    """
    def __init__(self, signal, obj, data, time):
        """
            signal: the 'signal' or 'name' for this Event [string]
            obj: the object emitting the Event [object]
            data: some piece of data relevant to the Event [object]
        """
        self.signal = signal
        self.object = obj or _NONE
        self.data = data
        self.time = time
        
class Callback(object):
    """
        Represents a callback
    """
    def __init__(self, function, time, args, kwargs):
        """
            @param function: the function to call
            @param time: the time this callback was added
        """
        self.valid = True
        self.wfunction = _getWeakRef(function, self.vanished)
        self.time = time
        self.args = args
        self.kwargs = kwargs

    def vanished(self, ref):
        self.valid = False
        

class EventManager(Logger):
    """
        Manages all Events
    """
    def __init__(self, use_logger=False, logger_filter=None):
        self.callbacks = {}
        self.use_logger = use_logger
        self.logger_filter = logger_filter

        # RLock is needed so that event callbacks can themselves send
        # synchronous events and add or remove callbacks
        self.lock = threading.RLock()

    def emit(self, signal, data=None, obj=None):
        """
            Emits an Event, calling any registered callbacks.

            event: the Event to emit [Event]
        """
        event = Event(signal, obj, data, time.time())
        
        with self.lock:
            callbacks = set()
            for tcall in set([_NONE, event.signal]):
                for ocall in set([_NONE, event.object]):
                    try:
                        callbacks.update(self.callbacks[tcall][ocall])
                    except KeyError:
                        pass

            # now call them
            for cb in callbacks:
                try:
                    if not cb.valid:
                        try:
                            self.callbacks[event.signal][event.object].remove(cb)
                        except (KeyError, ValueError):
                            pass
                    elif event.time >= cb.time:
                        if self.use_logger and (not self.logger_filter or \
                                re.search(self.logger_filter, event.signal)):
                                self.logdebug("Attempting to call "
                                    "%(function)s in response "
                                    "to %(event)s." % {
                                        'function': cb.wfunction(),
                                        'event': event.signal})

                        cb.wfunction().__call__(event.object, event.data, *cb.args, **cb.kwargs)
                            
                except Exception:
                    # something went wrong inside the function we're calling
                    traceback.print_exc(file=sys.stdout)
                    self.logdebug("Event callback exception caught!")
                    
        if self.use_logger:
            if not self.logger_filter or re.search(self.logger_filter,
                event.signal):
                self.logdebug("Sent '%(signal)s' event from "
                    "'%(object)s' with data '%(data)s'." %
                        {'signal' : event.signal, 'object' : repr(event.object),
                        'data' : repr(event.data)})

    def connect(self, signal, function, obj=None, *args, **kwargs):
        """
            Registers a callback.
            You should always specify at least one of signal or object.

            @param function: The function to call [function]
            @param signal: The 'signal' or 'name' of event to listen for. Defaults
                to any. [string]
            @param obj: The object to listen to events from. Defaults
                to any. [string]
        """
        
        with self.lock:
            # add the specified categories if needed.
            if not self.callbacks.has_key(signal):
                self.callbacks[signal] = weakref.WeakKeyDictionary()
            if obj is None:
                obj = _NONE
            try:
                callbacks = self.callbacks[signal][obj]
            except KeyError:
                callbacks = self.callbacks[signal][obj] = []

            # add the actual callback
            callbacks.append(Callback(function, time.time(), args, kwargs))

        if self.use_logger:
            if not self.logger_filter or re.search(self.logger_filter, signal):
                self.logdebug("Added callback %s for [%s, %s]" %
                        (function, signal, obj))

    def disconnect(self, signal, function, obj=None):
        """
            Unsets a callback

            The parameters must match those given when the callback was
            registered. (minus any additional args)
        """
        if obj is None:
            obj = _NONE
        remove = []

        with self.lock:
            try:
                callbacks = self.callbacks[signal][obj]
                for cb in callbacks:
                    if cb.wfunction() == function:
                        remove.append(cb)
            except KeyError:
                return
            except TypeError:
                return

            for cb in remove:
                callbacks.remove(cb)

        if self.use_logger:
            if not self.logger_filter or re.search(self.logger_filter, signal):
                self.logdebug("Removed callback %s for [%s, %s]" %
                        (function, signal, obj))

                
event_manager = EventManager()                

